#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup


setup(name='antichecked',
      version='1.0',
      description='Antichecked exceptions for Python.',
      author=None,
      author_email='SoniEx2 <endermoneymod@gmail.com>',
      url=None,
      py_modules=['antichecked'],
     )
